@@ -1,16 +1,20 @@
 import pandas as pd
 import numpy as np
-from eda_and_beyond.eda_tools import histograms_numeric_columns
 from .chart import plotly_heatmap, line_plotly
-from klib import cat_plot
 from operator import attrgetter
-from .utils import filter_non_capitalized_words_from_list
+from .utils import filter_non_capitalized_words_from_list, structure_concated_dataframe
+import warnings
+from .clean import clean_class
+from .const import std_vals
+import seaborn as sns
 
 
 class eda_class:
     def __init__(self, _df: pd.DataFrame, categorical_data: list = None) -> None:
         df = _df.copy()
+        self.na_like_items_in_str = std_vals.na_like_items_in_str
 
+        # Output dictionaries
         self.raw_input = df.copy()
         self.DTYPES = {}
         self.DOWNCASTING = {}
@@ -22,10 +26,10 @@ class eda_class:
 
         # Identifying dtypes
         self.DTYPES["boolean_like_columns"] = self.identify_boolean_like_columns()
+        self.DTYPES["datetime"] = self.identify_datetime_columns()
         self.DTYPES[
             "categorical_like_columns"
         ] = self.identify_categorical_like_columns()
-        self.DTYPES["datetime_like_columns"] = self.identify_datetime_like_columns()
         self.DTYPES["potential_uid_columns"] = self.find_unique_identifier_columns()
 
         # Interim calculations
@@ -37,6 +41,7 @@ class eda_class:
         )
 
         # Downcasting options
+        # Note: Int column cannot contain NaN values
         self.DOWNCASTING[
             "to_float32"
         ] = self.determine_candidates_for_float32_conversion()
@@ -64,9 +69,13 @@ class eda_class:
             _df=self.generate_correlation_of_missing_values(),
             title="Correlation of Missing Values",
         )
+        self.MISSING_VALUES[
+            "na_like_values_in_str_columns"
+        ] = self.get_mask_for_na_like_items_in_str_columns()
 
         # Duplicates
         self.DUPLICATES["redundant_columns"] = self.identify_redundant_columns()
+        self.DUPLICATES["duplicate_rows"] = self.identify_duplicate_rows()
 
         # Data analysis
         self.DATA_ANALYSIS["plotly_correlation_numerical"] = plotly_heatmap(
@@ -77,8 +86,7 @@ class eda_class:
             _df=self.correlation_matrix_non_numerical,
             title="Correlation Matrix Non Numerical",
         )
-        self.DATA_ANALYSIS["histogram_matplotlib"] = histograms_numeric_columns(
-            df=self.raw_input,
+        self.DATA_ANALYSIS["histogram_matplotlib"] = self.histograms_numeric_columns(
             numerical_columns=self.raw_input.select_dtypes(include="number"),
         ).figure
 
@@ -87,9 +95,6 @@ class eda_class:
             self.categorical_data = self.DTYPES["categorical_like_columns"]
         else:
             self.categorical_data = categorical_data
-        self.DATA_ANALYSIS["categorical_plot"] = cat_plot(
-            self.raw_input[self.categorical_data]
-        ).figure
         self.DATA_ANALYSIS["describe"] = self.generate_dataframe_describe(
             _df=self.raw_input
         )
@@ -97,13 +102,64 @@ class eda_class:
             "top_10_most_frequent_values"
         ] = self.identify_top_n_most_frequent_value_across_non_numeric_columns(n=10)
 
+    # Reference: eda_and_beyond
+    def histograms_numeric_columns(self, numerical_columns: list):
+        """
+        Args: dataframe, numerical columns (list)
+        Returns group histagrams
+        """
+        df = self.raw_input.copy()
+        f = pd.melt(df, value_vars=numerical_columns)
+        g = sns.FacetGrid(f, col="variable", col_wrap=4, sharex=False, sharey=False)
+        g = g.map(sns.distplot, "value")
+        return g
+
+    def get_mask_for_na_like_items_in_str_columns(self) -> pd.DataFrame:
+        # Select the requisite columns
+        df = self.raw_input.copy()
+        df = df.select_dtypes(exclude=np.number)
+        df = self._make_str_columns_lowercase(_df=df, cols=df.columns.to_list())
+        mask = df.isin(self.na_like_items_in_str)
+
+        return self.get_values_of_columns_based_on_mask(mask=mask)
+
+    def _make_str_columns_lowercase(
+        self, _df: pd.DataFrame, cols: list
+    ) -> pd.DataFrame:
+        df = _df.copy()
+        # Mixin function
+        clean_class_instance = clean_class()
+        return clean_class_instance._make_str_cols_lowercase(_df=df, cols=cols)
+
+    def get_values_of_columns_based_on_mask(self, mask: pd.DataFrame) -> pd.DataFrame:
+        # Select the requisite columns
+        df = self.raw_input
+        cols_in_mask = mask.columns.to_list()
+        df = df[cols_in_mask]
+
+        # Drop rows where all values are False
+        df = df[cols_in_mask]
+        df = df[mask].dropna(how="all")
+
+        # In the remaining lets record the unique values in a dict
+        dict_mask_values_in_df = {}
+        for col in df.columns:
+            dict_mask_values_in_df[col] = df[col].dropna().unique().tolist()
+
+        # Drop empty values from the dict
+        for key, value in dict_mask_values_in_df.copy().items():
+            if value == []:
+                dict_mask_values_in_df.pop(key, None)
+
+        return dict_mask_values_in_df
+
     def identify_if_dates_are_continuous(self) -> bool:
         _df = self.raw_input.copy()
         df_datetime_continuity = pd.DataFrame(
             columns=["col", "freq", "max_value_of_diff_between_periods"]
         )
 
-        for date_col in self.DTYPES["datetime_like_columns"]:
+        for date_col in self.DTYPES["datetime"]:
             df_temp = _df.copy()
             df_temp[date_col] = pd.to_datetime(df_temp[date_col])
 
@@ -128,7 +184,7 @@ class eda_class:
                     [df_datetime_continuity, df_datetime_continuity_temp], axis=0
                 )
 
-        return df_datetime_continuity
+        return structure_concated_dataframe(df_datetime_continuity)
 
     def generate_dataframe_describe(self, _df: pd.DataFrame) -> pd.DataFrame:
         df = _df.copy()
@@ -218,7 +274,7 @@ class eda_class:
         df_output = self.convert_float_to_percentage_in_dataframe(
             _df=df_output, cols=["percentage_missing_values"]
         )
-        return df_output
+        return structure_concated_dataframe(df_output)
 
     def generate_dataframe_with_percentage_of_non_missing_values(self) -> pd.DataFrame:
         df_output = self.generate_dataframe_with_percentage_of_missing_values(
@@ -228,7 +284,7 @@ class eda_class:
             columns={"percentage_missing_values": "percentage_non_missing_values"}
         )
         df_output = 1 - df_output
-        return df_output
+        return structure_concated_dataframe(df_output)
 
     def convert_float_to_percentage_in_dataframe(
         self, _df: pd.DataFrame, cols: list
@@ -289,11 +345,19 @@ class eda_class:
             len_post_dropna = len(df_temp)
 
             # Now check if the column is redundant
-            if len(df[col].unique()) == 1:
+            if len(df_temp[col].unique()) == 1:
                 redundant_columns_na_count[col] = len_pre_dropna - len_post_dropna
 
         return redundant_columns_na_count
 
+    def identify_duplicate_rows(self) -> pd.DataFrame:
+        df = self.raw_input.copy()
+        return (
+            df[df.duplicated(keep=False)]
+            .sort_values(by=df.columns.tolist())
+            .index.to_list()
+        )
+
     def make_upper_triangle_na(self, _df: pd.DataFrame) -> pd.DataFrame:
         df = _df.copy()
         mask = np.triu(np.ones_like(df, dtype=bool))
@@ -311,6 +375,11 @@ class eda_class:
             # Now check if the column is categorical like
             if len(df[col].unique()) > 2 and len(df[col].unique()) < 1 / 3 * len(df):
                 categorical_like_columns.append(col)
+
+        # Remove datetime columns from this
+        categorical_like_columns = list(
+            set(categorical_like_columns) - set(self.DTYPES["datetime"])
+        )
         return categorical_like_columns
 
     def make_diagonals_na(self, _df: pd.DataFrame) -> pd.DataFrame:
@@ -327,7 +396,7 @@ class eda_class:
         df_corr = self.make_upper_triangle_na(_df=df_corr)
         return df_corr
 
-    def identify_datetime_like_columns(self) -> pd.DataFrame:
+    def identify_datetime_columns(self) -> pd.DataFrame:
         df = self.raw_input.copy()
         df = df.apply(
             lambda col: pd.to_datetime(col, errors="ignore")
@@ -362,7 +431,7 @@ class eda_class:
         cols_uid = [
             "column_name",
             "unique_identifier_count",
-            "number_of_na",
+            "number_of_explicit_na",
             "number_of_duplicates",
         ]
         df_uid = pd.DataFrame(columns=cols_uid)
@@ -397,7 +466,7 @@ class eda_class:
                         ),
                     ]
                 )
-        return df_uid
+        return structure_concated_dataframe(_df=df_uid)
 
     def calculate_empirical_cdf(self, _df: pd.DataFrame, col: str) -> pd.DataFrame:
         df = _df[[col]].copy()
